#####################################################################
# CLASS TCP_pair
#####################################################################
#
# Contoller of TCP pairs sending data.
#
# VARIABLE:
#
# tcps, tcpr: sender TCP, receiver TCP 
# sn dn: source/dest node (only for setup_wnode)
# delay: delay between snode-sn/dnode-dn (only for setup_wnode)   
#
# pid: pair id, get from Agent_pair's pid (src->dst)
# cid: conn id, index of connection among each Agent_pair $pid (0,1,...,$var(conns_per_pair)-1)
# fid: flow id, global unique (given by $fid_) for each ($pid, $cid), also pass to $tcps (0,1,...,$pid*$cid-1)
#
# debug_mode: debug toggle, default value from global $var(debug_tcl); can be individully overloaded 
# ----------- e.g. $self instvar debug_mode; debugPrint $debug_mode 1 "msg"
#
# aggr_ctrl: Agent_pair for callback
# start_cbfunc: callback at start
# fin_cbfunc: callback at finish
#
# start_time: (memory) flow start time
# pkts: (memory) flow size
#
#
# FUNCTION:
#
# --- initializing ---
# init { args } ==> Agent/TCP/FullTcp set_callback{}
# setup { snode dnode }
# setup_wnode { snode dnode linkdelay }
#
# --- helper ---
# setpid { pairid } 
# setconnid { connid }
# setfid { flowid } 
# set_callback { controller startfunc finfunc }
# set_debug_mode { mode }
#
# --- running ---
# start { npkts taskid isstable } ==> start_notify{}
# stop {}
#
# --- callback ---
# start_notify {} ==> Agent_pair start_notify{}
# fin_notify {} ==> Agent_pair fin_notify{}
#
# --- Agent/TCP/FullTcp ---
# set_callback { tcp_pair }
# done_data {} ==> fin_notify{}
#
#####################################################################

Class TCP_pair

#:::::::::::::::: initializing ::::::::::::::::

TCP_pair instproc init { args } {
    eval $self next $args

    global myAgent var
    $self instvar tcps tcpr
    $self instvar pid cid fid 
    $self instvar debug_mode

    $self set tcps [new $myAgent] 
    $self set tcpr [new $myAgent]

    $self set pid 0
    $self set cid 0
    $self set fid 0

    $self set debug_mode $var(debug_tcl)

    $tcps set_callback $self
}

TCP_pair instproc setup { snode dnode } {
    # PUBLIC
    # connect agents directly to $snode/$dnode for faster simulation.
    
    global ns
    $self instvar tcps tcpr

    $ns attach-agent $snode $tcps;
    $ns attach-agent $dnode $tcpr;

    $tcpr listen
    $ns connect $tcps $tcpr
}


TCP_pair instproc setup_wnode { snode dnode linkdelay } {
    # PUBLIC
    # create new nodes $sn/$dn for agents, and connect to $snode/$dnode with link with $linkdelay.
    # CAUTION: if the number of pairs is large, simulation gets way too slow, and memory consumption gets very very large.. use "setup" if possible in such cases.

    global ns var
    $self instvar tcps tcpr
    $self instvar sn dn
    $self instvar delay

    $self set sn [$ns node]
    $self set dn [$ns node]
    $self set delay linkdelay

    $ns duplex-link $snode $sn  [set $var(link_rate)]Gb $delay  DropTail
    $ns duplex-link $dn $dnode  [set $var(link_rate)]Gb $delay  DropTail

    $ns attach-agent $sn $tcps;
    $ns attach-agent $dn $tcpr;

    $tcpr listen
    $ns connect $tcps $tcpr
}

#:::::::::::::::: helper ::::::::::::::::

TCP_pair instproc setpid { pairid } {
    $self instvar pid
    $self set pid $pairid
}

TCP_pair instproc setconnid { connid } {
    $self instvar cid
    $self set cid $connid
}

TCP_pair instproc setfid { flowid } {
    $self instvar tcps tcpr
    $self instvar fid
    $self set fid $flowid
    $tcps set fid_ $fid;
    $tcpr set fid_ $fid;
}

TCP_pair instproc set_callback { controller startfunc finfunc } {
    $self instvar aggr_ctrl start_cbfunc fin_cbfunc
    $self set aggr_ctrl $controller
    $self set start_cbfunc $startfunc
    $self set fin_cbfunc $finfunc
}

# p2-task (GONG)
# TCP_pair instproc set_callback_ext { aggrctrl aggrstartfunc aggrfinfunc taskctrl taskgenfunc taskfinfunc } {
#     $self instvar aggr_ctrl task_ctrl aggr_start_cbfunc aggr_fin_cbfunc task_gen_cbfunc task_fin_cbfunc
    
#     $self set task_ctrl $taskctrl
#     $self set task_gen_cbfunc $taskgenfunc
#     $self set task_fin_cbfunc $taskfinfunc
#     $self set aggr_ctrl $aggrctrl
#     $self set aggr_start_cbfunc $aggrstartfunc
#     $self set aggr_fin_cbfunc $aggrfinfunc
# }

TCP_pair instproc set_debug_mode { mode } {
    $self instvar debug_mode
    $self set debug_mode $mode
}

#:::::::::::::::: running ::::::::::::::::

TCP_pair instproc start { npkts taskid isstable } { 
    # PUBLIC
    # start transmission of $npkts * pkt_size_(1460) bytes, increase $flow_gen

    global ns var flow_gen
    $self instvar tcps tcpr
    $self instvar start_time pkts

    $self set start_time [$ns now]
    $self set pkts $npkts

    # p2-task (GONG)
    $self instvar task_id
    $self set task_id $taskid
    
    # load stable (GONG)
    $self instvar stable
    $self set stable $isstable

    # p2-task (GONG) comment this out
    # if { $flow_gen >= $var(sim_end) } {
    #    return
    # }
    incr flow_gen

    $self start_notify

    $tcpr set flow_remaining_ [expr $npkts * $var(pkt_size_)]
    $tcps set task_id_ $taskid   ; # p2-task (GONG)
    $tcpr set task_id_ $taskid   ; # p2-task (GONG) needed ?
    $tcps set signal_on_empty_ TRUE
    $tcps advance-bytes [expr $npkts * $var(pkt_size_)]
}

# p1-flow-based task (GONG)
# TCP_pair instproc start_ext { nbytes tagname taskprio } {

#     global ns
#     $self instvar tcps tcpr
#     $self instvar start_time pkts
#     $self instvar tag task_prio

#     $self set start_time [$ns now]
#     $self set pkts [expr $nbytes / $var(pkt_size_)]
#     $self set tag $tagname
#     $self set task_prio $taskprio

#     $self start_notify

#     $tcpr set flow_remaining_ [expr $nbytes]
#     $tcps set task_id_ $task_prio
#     $tcpr set task_id_ $task_prio  ; # needed ?
#     $tcps set signal_on_empty_ TRUE
#     $tcps advance-bytes [expr $nbytes]
# }

TCP_pair instproc stop {} {
    # PUBLIC
    # stop tcps/tcpr between each transmission (non-persistent connection)

    $self instvar tcps tcpr

    $tcps reset
    $tcpr reset
}

#:::::::::::::::: callback ::::::::::::::::

TCP_pair instproc start_notify {} {
    # CALLBACK
    # called by TCP_pair start{}, call its Agent_pair start_notify{}

    $self instvar pid cid
    $self instvar aggr_ctrl start_cbfunc task_ctrl task_gen_cbfunc
    $self instvar pkts tag
    
    # p2-task (GONG)
    $self instvar task_id
    
    # load stable (GONG)
    $self instvar stable
    
    # p2-task (GONG)
    if { [info exists task_ctrl] } {
        $task_ctrl $task_gen_cbfunc $pid $cid $tag $pkts
    } elseif { [info exists aggr_ctrl] } {
        # $aggr_ctrl $start_cbfunc $cid $pkts $task_id
        $aggr_ctrl $start_cbfunc $cid $pkts $task_id $stable ; # load stable (GONG)
    }

    
}

TCP_pair instproc fin_notify {} {
    # CALLBACK
    # called by Agent/TCP/FullTcp done_data{}, call its Agent_pair fin_notify{}

    global ns
    $self instvar pid cid
    $self instvar aggr_ctrl fin_cbfunc 
    $self instvar task_ctrl aggr_fin_cbfunc task_fin_cbfunc
    $self instvar start_time
    $self instvar tag pkts
    
    # p2-task (GONG)
    $self instvar task_id
    
    # load stable (GONG)
    $self instvar stable

    # p2-task (GONG)
    if { [info exists task_ctrl] } {
        $task_ctrl $task_fin_cbfunc $pid $cid $tag $pkts [expr [$ns now] - $start_time]
        $aggr_ctrl $aggr_fin_cbfunc $cid
    } elseif { [info exists aggr_ctrl] } {
        # $aggr_ctrl $fin_cbfunc $cid [expr [$ns now] - $start_time] $task_id
        $aggr_ctrl $fin_cbfunc $cid [expr [$ns now] - $start_time] $task_id $stable ; # load stable (GONG)
    }
}

# Agent/TCP/FullTcp callback
Agent/TCP/FullTcp instproc set_callback { tcp_pair } {
    $self instvar ctrl
    $self set ctrl $tcp_pair
}

Agent/TCP/FullTcp instproc done_data {} {
    # called by FullTcpAgent::bufferempty() when the sender buffer is empty
    $self instvar ctrl
    if { [info exists ctrl] } {
       $ctrl fin_notify
    }
}
